;---------------------------------------------------------------
; RetroForth/Native 7.4
;---------------------------------------------------------------
include "macros.fasm"

use32
ORG 0x11000
MAIN:	call init_idt
	call init_screen
	call init_retroforth
	lidt [idt_pointer]
	sti
	jmp  interpret
;---------------------------------------------------------------

code 'emit', emit
	cmp	al,8
	 je .backspace
	cmp	al,9
	 je .tab
	cmp	al,10
	 je .cr

	push	edi	        ; printable character
	mov	ah,byte [attr]
	movzx	edi, word [xy]
	add	edi, [vfb]
	stosw
	add	word [xy], 2
	call	scroll
	call	update_cursor
	pop	edi
	drop
	ret
.backspace:	push edi
	sub word [xy], 2
	movzx edi, word [xy]
	add edi, [vfb]
	xor al, al
	stosb
	call update_cursor
	pop edi
	drop
	ret
.tab:
	and dword [xy], $FFFFFFFFFFFFFFF0 ;~0x0F            ;; round [xy] up to the nearest 16
	add dword [xy], 0x10
	call update_cursor
	drop
	ret
.cr:	call cr
	drop
	ret
next

init_screen:
;
; Get cursor position from BIOS data area
;
	movzx	eax,word [0x44A]        ;columns
	mov     [vw],eax
	movzx	eax,byte [0x484]        ;rows-1
	inc	al
	mov     [vh],eax
	movzx	eax,byte [0x451]        ;cursor row
	mul	dword [vw]
	movzx	edx,byte [0x450]        ;cursor column
	add	eax,edx
	shl	eax,1
	mov     [xy],eax
	ret

;---------------------------------------------------------------
; ( KEY, a generic keyboard driver )
code 'key', key
	dup			        ; key
	  xor eax,eax	        ;  clear eax
.a:	 in al,64h		 ;  Is any data waiting?
	  test al,1		        ;  Is character = ASCII 0?
	  jz .a 	        ;  Yes? Try again
	in al,60h	        ;  Otherwise, read scancode
	xor edx,edx	        ;  edx: 0=make, 1=break
	test al,80h	        ;  Is character = HEX 80h?
	jz .b		        ;  Skip the next line
	inc edx 	        ;  Update edx
.b:	 and al,7Fh		 ;  Filters to handle
	cmp al,39h	        ;  the ignored keys
	  ja .a 	        ;  We just try another key
	mov ecx,[board]         ;  Load the keymap
	mov al,[ecx+eax]        ;  Get the key ASCII char
	  or al,al		        ;  Is is = 0?
	js .shift		        ;  No, use CAPITALS
	jz .a		        ;  Ignore 0's
	or dl,dl		        ;  Filter for break code
	jnz .a		        ;  Ignore break code
	dup
	call emit
	  ret
.shift:  mov ecx,[edx*4 + .shifts]	 ;  Load the CAPITAL keymap
	mov [board],ecx 	        ;  Store into BOARD pointer
	jmp .a			  ;  And try again
.shifts dd shift,alpha
board dd alpha
alpha:
  db 0,27,"1234567890-=",8	        ;00-0E
  db 9,"qwertyuiop[]",10	        ;0F-1C
  db 0,"asdfghjkl;'`"		        ;1D-29
  db -1,"\zxcvbnm,./",-1,"*",0,32,-2    ;2A-3A
shift:
  db 0,27,"!@#$%^&*()_+",8	        ;00-0E
  db 9,"QWERTYUIOP{}",10	        ;0F-1C
  db 0,'ASDFGHJKL:"~'		        ;1D-29
  db -1,"|ZXCVBNM<>?",-1,"*",0,32,-2    ;2A-3A
;---------------------------------------------------------------
vfb	dd      0xB8000 	        ;video frame buffer (text/gfx)
vw	dd      80
vh	dd      25
xy	dd      0
attr	db      15

;---------------------------------------------------------------
update_cursor:
;
; Physically move the cursor to [x,y], using port I/O
;
; Index port: 0x3d4
; Data port:  0x3d5
;
	push	eax, edx
	mov	al, 0x0f        ;Index reg.: cursor pos (low byte)
	mov	dx, 0x3d4
	out	dx, al
	mov	ax, word [xy]	     ;Write low byte
	shr	ax, 1
	inc	edx	        ;0x3d5
	out	dx, al
	mov	al, 0x0e        ;Index reg.: cursor pos (high byte)
	dec	edx	        ;0x3d4
	out	dx, al
	mov	al, ah	        ;Write high byte
	inc	edx	        ;0x3d5
	out	dx, al
	pop	eax, edx
	ret
;---------------------------------------------------------------
cr:
	push	ecx,edx
	mov	ax,word [xy]
	xor	dx,dx
	mov	cx,word [vw]
	shl	cx,1
	div	cx	        ;dx = remainder
	mov	ax,word [vw]
	shl	ax,1
	sub	ax,dx
	add	word [xy],ax
	call	scroll
	call	update_cursor
	pop	ecx,edx
	ret
;---------------------------------------------------------------
scroll:
;
; If the cursor is past the bottom of the screen, scroll down 1 line.
;
	mov	eax,[vw]
	imul	eax,[vh]
	shl	eax,1
	cmp	word [xy],ax		     ;need to scroll?
	jb	.end

	pusha
	mov	edi,[vfb]	        ;Move rows up 1
	mov	esi,[vw]
	shl	esi,1
	add	esi,edi
	shr	eax,1
	mov	ecx,eax
	sub	ecx,[vw]
	cld
	rep	movsw

	mov	ah,[attr]	        ;Clear bottom row
	xor	al,al
	mov	ecx,[vw]
	rep	stosw

	mov	eax,[vw]		        ;move cursor
	shl	eax,1
	sub     [xy],eax
	popa
.end:	ret

include 'idt.inc'
include 'hdd.inc'

include "retroforth.asm"
